@import url("resource:///plugins/glade/themes/shared.css");

gbpgladeproperties {
  font-size: 0.83333em;
}

gbpgladeproperties entry,
gbpgladeproperties spinbutton {
  min-height: 24px;
}

gbpgladeproperties checkbutton check {
  min-width: 10px;
  min-height: 10px;
}

gbpgladeproperties spinbutton button {
  min-height: 0px;
  min-width: 0px;
  padding: 0 4px;
}

gbpgladeproperties spinbutton entry {
  min-height: 0px;
  min-width: 0px;
  padding: 0 4px;
}

gbpgladeproperties switch slider {
  min-height: 16px;
  min-width: 18px;
}

gbpgladeproperties button.combo {
  min-height: 18px;
  padding: 2px;
}

.glade-chooser button {
  min-height: 12px;
  min-width: 12px;
}

.glade-chooser {
  font-size: 0.8333em;
  padding: 6px;
}
