@import url("resource:///org/gnome/builder/themes/shared.css");

ideframeheader > button:last-child {
  margin-right: 3px;
}

entry.search-missing {
  background-color: #cc0000;
  color: white;
  text-shadow: none;
}

entry.search-missing > image {
  color: white;
}


/* tweak icons for treeviews */
treeview.image { color: alpha(currentColor, 0.8); }
treeview.image:selected { color: alpha(@theme_selected_fg_color, 0.9); }


/* utilities stack switcher */
ideeditorutilities > dzldockpaned > box > stackswitcher {
  margin: 6px;
}


/* buildui panel */
ideeditorsidebar notebook header {
  background: transparent;
}
ideeditorsidebar notebook header tab {
  padding: 0;
}

ideeditorproperties entry:last-child {
  border-radius: 0;
  border-right: none;
  border-left: none;
}

/* Omnibar */
popover.omnibar {
  padding: 0;
}
popover.omnibar list {
  border-top: 1px solid @borders;
  border-radius: 0 0 5px 5px;
}
popover.omnibar list row:not(:last-child) {
  border-bottom: 1px solid alpha(@borders, 0.3);
}
popover.omnibar row.notification .title {
  font-weight: bold;
}
popover.omnibar row.notification .body {
  opacity: 0.8;
  font-size: 0.95em;
}

/* Notifications button (transfers, etc) */
popover.notificationsbutton {
  padding: 0;
}
popover.notificationsbutton list {
  background: transparent;
  border-radius: 5px;
}
popover.notificationsbutton list row:not(:last-child) {
  border-bottom: 1px solid alpha(@borders, 0.3);
}
popover.notificationsbutton row.notification .title {
  font-weight: bold;
}
popover.notificationsbutton row.notification .body {
  opacity: 0.8;
  font-size: 0.95em;
}
popover.notificationsbutton button.circular {
  min-height: 24px;
  min-width: 24px;
}

/* development styles */
window.development-version  headerbar:last-child {
  background: transparent -gtk-icontheme("system-run-symbolic") 80% 0/128px 128px no-repeat,
              linear-gradient(to left,
                              mix(@theme_fg_color, @theme_bg_color, 0.5) 0%,
                              @theme_bg_color 25%);
  color: alpha(@theme_fg_color, 0.2);
}

window.development-version headerbar label:not(:disabled) {
  color: @theme_fg_color;
}

window.development-version headerbar .suggested-action label {
  color: @theme_selected_fg_color;
}


/* preferences overrides */
dzlpreferencesgroup frame,
dzlpreferencesgroup .frame {
  background: @content_view_bg;
  border-radius: 9px;
}
dzlpreferencesgroup frame border,
dzlpreferencesgroup .frame border {
  border-radius: 9px;
}
dzlpreferencesgroup list {
  background: transparent;
}
dzlpreferencesgroup list row {
  min-height: 28px;
}
dzlpreferencesgroup list row:first-child {
  border-top-left-radius: 9px;
  border-top-right-radius: 9px;
}
dzlpreferencesgroup list row:last-child {
  border-bottom-left-radius: 9px;
  border-bottom-right-radius: 9px;
}
