popover.hoverer {
  padding: 0px;
}

popover.hoverer > box > :not(:first-child) {
  margin-top: 8px;
}

popover.hoverer .title {
  font-weight: 700;
  font-size: 10pt;
  margin-bottom: 6px;
}

popover.hoverer box.linked button {
  padding: 0px 6px;
  font-size: 0.9em;
}

.hoverer-box {
  padding: 12px;
}

.hoverer-box:not(:last-child) {
  border-bottom: 1px solid @borders;
}
